
/******************************************************************************
* MODULE     : analytic_eval.hpp
* DESCRIPTION: Evaluation of an analytic functions in a point
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "analytic.hpp"
namespace mml {

/******************************************************************************
* Evaluation in double precision case
******************************************************************************/

template<> double_analytic::complex_type
analytic_rep<double_analytic>::evaluate
  (const double_analytic::complex_type& z) const
{
  double_analytic::series_type   s= germ ();
  double_analytic::positive_type r= double_analytic::_abs (z);
  double_analytic::positive_type R= radius_between (r, radius ());
  //cout << "[" << r << ", " << R << "]";
  double_analytic::positive_type M= bound (R);
  double_analytic::positive_type a= r / R;
  double_analytic::positive_type B= M / (1.0 - a);
  double_analytic::positive_type C= 2.625223256644e-33; // 2^(-52)
  double_analytic::complex_type sum (0.0), pow (1.0);
  for (nat i=0; i<1000; i++) {
    sum += s[i] * pow;
    pow *= z;
    B   *= a;
    if (B < C * abs(sum)) break;
  }
  return sum;
}

/******************************************************************************
* Evaluation in certified precision case
******************************************************************************/

class evaluate_analytic_rep: public complex_rep {
private:
  analytic<exact_analytic> f;  // the function being evaluated
  complex z;                   // the point where we evaluate

public:
  inline evaluate_analytic_rep (const analytic<exact_analytic>& f2,
				const complex& z2):
    f (f2), z (z2) { init (); }
  generic expression () const { return f->expression (as_generic (z)); }
  complex_ball compute () {
    exact_analytic::series_type s= f->germ ();
    //cout << "prec= " << prec << "\n";
    //cout << "<" << prec << ": " << f->complexity << ">"; cout.flush ();
    //cout << "[" << mml_used_bytes () << "]"; cout.flush ();
    // Error estimate for tail of series
    real_ball r = exact_analytic::_abs (z);
    real_ball R = radius_between (r, f->radius ());
    real_ball M = f->bound (R);
    real_ball a = r / R;
    //cout << "[[" << M << "," << a << "]]";
    //cout << "M= " << M << "\n";
    //cout << "a= " << a << "\n";

    nat i;
    real_ball err= M;
    //complex sum= 0;
    complex prod= 1;
    complex_ball_radius rad= 0;
    complex_ball ret= 0;
    for (i=0; true; i++) {
      complex term= s[i] * prod;
      //cout << "[" << i << "]"; cout.flush ();
      //cout << "\t" << i << "\t" << approx_raw (term, prec) << "\n";
      //sum  += term;
      prod *= z;
      err  *= a;
      rad= max (rad, radius (term->compute (prec)));
      ret += current (term); // FIXME: improve me
      /*
      if (i%10 == 0) {
	cout << "\t\t" << i << "\t" << term->compute (prec) << "\n";
	cout << "\t\t\t" << ret << " [" << mml_default_prec << "]" << "\n";
      }
      */
      if (rad * complex_ball_radius (1e-6) > bnd_U (err)) break;
    }
    complex_ball back= ret + complex_ball (0, bnd_U (err));
    //cout << "[" << i << "]"; cout.flush ();
    //cout << "[#" << mml_used_bytes () << "]"; cout.flush ();
    //cout << "Done\n";
    //cout << f->complexity << "\t" << prec << "\t" << i << "\t" << back << "\n";
    //cout << "\t" << radius (Re (back)) << "\t" << err << "\n";

    /*
    nat n= i;
    cout << "---------- " << f->complexity << " -----------] " << prec << "\n";
    for (i=0; i<min(n,(nat)10); i++)
      cout << "  " << i << "\t" << max (Re (f[i])->prec, Im (f[i])->prec)
	   << "\t" << approx_current (f[i]) << "\n";
    cout << "..... -> " << n << "\n";
    */
    return back;
  }
};

complex
evaluate_analytic (const analytic<exact_analytic>& f, const complex& z) {
  return (complex_rep*) new evaluate_analytic_rep (f, z);
}

template<> exact_analytic::complex_type
analytic_rep<exact_analytic>::evaluate
  (const exact_analytic::complex_type& z) const
{
  return evaluate_analytic (analytic<exact_analytic> (this, true), z);
}

} // namespace mml
