
/******************************************************************************
* MODULE     : iterator.hpp
* DESCRIPTION: Iterators
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_ITERATOR_HPP
#define __MML_ITERATOR_HPP
#include "base.hpp"
#include "operators.hpp"
namespace mml {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
#define Iterator iterator<C>
#define Iterator_rep iterator_rep<C>
TMPL class iterator;
TMPL inline bool busy (const Iterator& it);
TMPL inline bool done (const Iterator& it);
TMPL inline bool init (const Iterator& it);

/******************************************************************************
* The abstract iterator class
******************************************************************************/

TMPL_DEF
class iterator_rep REP_STRUCT {
protected:
  virtual bool is_busy () = 0;
  virtual bool is_done () { return !is_busy (); }
  virtual bool is_init () {
    assert (false); /* not implemented */ return false; }
  virtual void advance () = 0;
  virtual void regress () {
    assert (false); /* not implemented */ }
  virtual C current () = 0;
  virtual Iterator_rep* clone () {
    assert (false); /* not implemented */ return NULL; }

public:
  iterator_rep () {}
  virtual ~iterator_rep () {}
  friend class Iterator;
  friend bool busy LESSGTR (const Iterator& it);
  friend bool done LESSGTR (const Iterator& it);
  friend bool init LESSGTR (const Iterator& it);
};

TMPL_DEF
class iterator {
INDIRECT_PROTO_1 (iterator, iterator_rep, C)
public:
  iterator ();
  iterator (const C& c);
  inline C operator * () const { return rep->current(); }
  const Iterator& operator ++ () const {
    if (rep->ref_count>1) {
      rep->ref_count--;
      *(const_cast<Iterator_rep**> (&rep))= rep->clone ();
    }
    rep->advance(); return *this; }
  const Iterator& operator -- () const {
    if (rep->ref_count>1) {
      rep->ref_count--;
      *(const_cast<Iterator_rep**> (&rep))= rep->clone ();
    }
    rep->regress(); return *this; }
  friend bool busy LESSGTR (const Iterator& it);
  friend bool done LESSGTR (const Iterator& it);
  friend bool init LESSGTR (const Iterator& it);
};
INDIRECT_IMPL_1 (iterator, iterator_rep, typename C, C)

TMPL inline bool busy (const Iterator& it) { return it.rep->is_busy (); }
TMPL inline bool done (const Iterator& it) { return it.rep->is_done (); }
TMPL inline bool init (const Iterator& it) { return it.rep->is_init (); }

/******************************************************************************
* Output
******************************************************************************/

TMPL generic
as_generic (const Iterator& it) {
  generic args= comma ();
  for (; busy (it); ++it)
    args= comma (args, as_generic (*it));
  return apply ("iterator", args);
}

TMPL ostream&
operator << (ostream& out, const Iterator& it) {
  if (done (it)) out << "iterator (";
  else {
    out << "iterator (" << *it; ++it;
    for (; busy (it); ++it)
      out << ", " << *it;
  }
  return out<< ")";
}

/******************************************************************************
* Finite iterators
******************************************************************************/

TMPL_DEF
class finite_iterator_rep: public Iterator_rep {
  C*  a;  // the items
  nat n;  // the number of items
  nat i;  // how many items have been processed

public:
  finite_iterator_rep (C* a2, nat n2): a(a2), n(n2), i(0) {}
  ~finite_iterator_rep () { delete[] a; }

protected:
  bool is_busy () { return i<n; }
  bool is_init () { return i==0; }
  void advance () { i++; }
  void regress () { i--; }
  C current () { return a[i]; }
  Iterator_rep* clone () {
    finite_iterator_rep* rep= new finite_iterator_rep (a, n);
    rep->i= i;
    return rep;
  }
};

TMPL
Iterator::iterator () {
  C* a= new C[0];
  rep= new finite_iterator_rep<C> (a, 0);
}

TMPL
Iterator::iterator (const C& c) {
  C* a= new C[0];
  a[0]= c;
  rep= new finite_iterator_rep<C> (a, 1);
}

TMPL Iterator
seq (const C& c1) {
  C* a= new C[1];
  a[0]= c1;
  return Iterator (new finite_iterator_rep<C> (a, 1));
}

TMPL Iterator
seq (const C& c1, const C& c2) {
  C* a= new C[2];
  a[0]= c1; a[1]= c2;
  return Iterator (new finite_iterator_rep<C> (a, 2));
}

TMPL Iterator
seq (const C& c1, const C& c2, const C& c3) {
  C* a= new C[3];
  a[0]= c1; a[1]= c2; a[2]= c3;
  return Iterator (new finite_iterator_rep<C> (a, 3));
}

TMPL Iterator
seq (const C& c1, const C& c2, const C& c3,
     const C& c4)
{
  C* a= new C[4];
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4;
  return Iterator (new finite_iterator_rep<C> (a, 4));
}

TMPL Iterator
seq (const C& c1, const C& c2, const C& c3,
     const C& c4, const C& c5)
{
  C* a= new C[4];
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4; a[4]= c5;
  return Iterator (new finite_iterator_rep<C> (a, 5));
}

TMPL Iterator
seq (const C& c1, const C& c2, const C& c3,
     const C& c4, const C& c5, const C& c6)
{
  C* a= new C[4];
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4; a[4]= c5; a[5]= c6;
  return Iterator (new finite_iterator_rep<C> (a, 6));
}

/******************************************************************************
* Joining iterators
******************************************************************************/

TMPL_DEF
class join_iterator_rep: public Iterator_rep {
  Iterator it1, it2;

public:
  join_iterator_rep (const Iterator it1b, const Iterator it2b):
    it1(it1b), it2(it2b) {}

protected:
  bool is_busy () { return busy (it1) || busy (it2); }
  void advance () { if (busy (it1)) ++it1; else ++it2; }
  bool is_init () { return init (it1) && init (it2); }
  void regress () { if (init (it2)) --it1; else --it2; }
  C current () { return busy (it1)? *it1: *it2; }
  Iterator_rep* clone () {
    return new join_iterator_rep (it1, it2); }
};

TMPL Iterator
operator * (const Iterator& it1, const Iterator& it2) {
  return Iterator (new join_iterator_rep<C> (it1, it2));
}

/******************************************************************************
* Big operators on iterators
******************************************************************************/

template<typename Op, typename C> C
big (const Iterator& it) {
  if (done (it)) return Op::template neutral<C> ();
  C r= *it;
  for (++it; busy (it); ++it)
    Op::in_place (r, *it);
  return r;
}

TMPL inline C big_add (const Iterator& it) { return big<add_op> (it); }
TMPL inline C big_mul (const Iterator& it) { return big<mul_op> (it); }
TMPL inline C big_or  (const Iterator& it) { return big<or_op > (it); }
TMPL inline C big_and (const Iterator& it) { return big<and_op> (it); }
TMPL inline C big_min (const Iterator& it) { return big<min_op> (it); }
TMPL inline C big_max (const Iterator& it) { return big<max_op> (it); }

#undef TMPL_DEF
#undef TMPL
#undef Iterator
#undef Iterator_rep
} // namespace mml
#endif // __MML_ITERATOR_HPP
