
/******************************************************************************
* MODULE     : string_scnot.cpp
* DESCRIPTION: Scientific notation for intervals
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "string.hpp"
#include "integer.hpp"
namespace mml {

static void
decompose (const string& s, int& sign, string& mant, integer& expo) {
  // Sign
  if (N(s) > 0 && (s[0] == '+' || s[0] == '-')) {
    decompose (s (1, N(s)), sign, mant, expo);
    if (s[0] == '-') sign= -sign;
    return;
  }

  // Special values
  sign= 1; mant= ""; expo= 0;
  if (locase (s) == "nan") { mant= "NaN"; return; }
  if (locase (s) == "fuzz") { mant= "Fuzz"; return; }
  if (locase (s) == "infty") { mant= "Infty"; return; }
  if (locase (s) == "infinity") { mant= "Infty"; return; }

  // Get mantissa and decimal point
  int dot= -1;
  nat i= 0, n= N(s);
  while (i<n && (s[i] == '.' || (s[i] >= '0' && s[i] <= '9'))) {
    if (s[i] == '.') {
      if (dot != -1) { mant= "NaN"; return; }
      dot= i++;
    }
    else mant << s[i++];
  }
  if (dot == -1) dot= i;

  // Get exponent
  if (i<n && (s[i] == 'e' || s[i] == 'E')) {
    i++;
    int expo_sign= 1;
    while (i<n && (s[i] == '+' || s[i] == '-')) {
      if (s[i] == '-') expo_sign= -expo_sign;
      i++;
    }
    while (i<n && s[i] >= '0' && s[i] <= '9') {
      expo= 10*expo + ((int) (s[i] - '0'));
      i++;
    }
    expo= expo_sign * expo;
  }

  // Finish
  if (i != n) { mant= "NaN"; return; }
  if (mant == "") { mant= "0"; dot++; }
  expo += (dot - 1);
  while (N(mant) > 1 && mant[0] == '0') {
    mant= mant (1, N(mant));
    expo= expo - 1;
  }
  if (s == "0") sign= 0;
}

static string
recompose (int sign, const string& mant, const integer& i) {
  if (sign == 0) return "0";
  if (locase (mant) == "nan") return "NaN";
  if (mant == "0") return "0e" * as_string (i);
  if (sign < 0) return "-" * recompose (1, mant, i);
  if (locase (mant) == "infty") return "Infty";
  if (i == -1) return "0." * mant;
  if (i == -2) return "0.0" * mant;
  if (i == -3) return "0.00" * mant;
  if (N(mant) == 1) return mant * "e" * as_string (i);
  if (i >= 0 && i<6 && N(mant)>i+1)
    return mant (0, as_int (i)+1) * "." * mant (as_int (i)+1, N(mant));
  return mant (0, 1) * "." * mant (1, N(mant)) * "e" * as_string (i);
}

static void
inc_mantissa (string& mant, integer& expo) {
  string mant2= as_string (integer (mant) + 1);
  if (N (mant2) > N (mant)) expo= expo + 1;
  mant= mant2;
}

static string
make_range (string& mant1, integer& expo1,
	    string& mant2, integer& expo2)
{
  if (expo2 == expo1 + 1) {
    mant1= "0" * mant1;
    expo1= expo1 + 1;
  }
  if (expo2 != expo1) return "0e" * as_string (expo2 + 1);

  nat i, n= min (N(mant1), N(mant2));
  for (i=0; i<n; i++)
    if (mant1[i] != mant2[i]) {
      if (mant2[i] == mant1[i] + 1) {
	i++;
	while (i<n && (mant1[i] == '9' && mant2[i] == '0')) i++;
	if (i >= n ||
	    (mant1[i] >= '0' && mant1[i] <= '4') ||
	    (mant2[i] >= '5' && mant2[i] <= '9')) i--;
      }
      break;
    }
  string mant2b= mant2 (0, i);
  if (i>0 && mant1[i-1] == mant2[i-1]) {
    if (i<N(mant1) && mant1[i] >= '5')
      inc_mantissa (mant2b, expo2);
    else if (i<N(mant2) && mant2[i] >= '5') {
      bool inc= (i != 0 && mant2[i-1] >= '5');
      mant2b= mant2b (0, N(mant2b) - 1);
      if (inc) inc_mantissa (mant2b, expo2);
    }
  }
  if (mant2b == "") return "0e" * as_string (expo2 + 1);
  else return recompose (1, mant2b, expo2);
}

string
make_range (const string& s1, const string& s2) {
  int     sign1, sign2;
  string  mant1, mant2;
  integer expo1, expo2;
  decompose (s1, sign1, mant1, expo1);
  decompose (s2, sign2, mant2, expo2);
  if (locase (mant1) == "nan" || locase (mant2) == "nan") return "NaN";
  if (locase (mant1) == "infty" || locase (mant2) == "infty") {
    if (mant1 == mant2 && sign1 == sign2)
      return recompose (sign1, mant1, expo1);
    return "Fuzz";
  }
  if (sign1 != sign2) {
    integer expo= expo1;
    if (sign1 == 0 || (sign2 != 0 && expo2 > expo1)) expo= expo2;
    return "0e" * as_string (expo+1);
  }
  if (sign1 == 0) return "0";
  if (sign1 < 0) return "-" * make_range (mant2, expo2, mant1, expo1);
  else return make_range (mant1, expo1, mant2, expo2);
}

void
decompose_range (const string& s, string& l, string& r) {
  int     sign;
  string  mant;
  integer expo;
  decompose (s, sign, mant, expo);
  if (locase (mant) == "fuzz") { l= "-Infty"; r= "Infty"; return; }
  if (locase (mant) == "nan") { l= "NaN"; r= "NaN"; return; }
  if (locase (mant) == "infty") { l= r= recompose (sign, mant, 0); return; }
  if (sign == 0) { l= r= "0"; }
  if (mant == "0") {
    l= recompose (-1, "1", expo);
    r= recompose (1, "1", expo);
  }
  else {
    string mant1= as_string (integer (mant) - 1) * "5";
    string mant2= mant * "5";
    l= recompose (sign, sign>0? mant1: mant2, expo);
    r= recompose (sign, sign>0? mant2: mant1, expo);
  }
}

string
trunc_digits (const string& s, nat dd) {
  if (locase (s) == "nan") return "NaN";
  if (locase (s) == "infty") return "Infty";
  if (locase (s) == "-infty") return "-Infty";
  if (locase (s) == "fuzz") return "Fuzz";
  int     sign;
  string  mant;
  integer expo;
  decompose (s, sign, mant, expo);
  if (dd == 0 || N(mant) <= dd) return recompose (sign, mant, expo);
  string mant2= mant (0, dd);
  if (mant[dd] >= '5') inc_mantissa (mant2, expo);
  return recompose (sign, mant2, expo);
}

} // namespace mml
