
/******************************************************************************
* MODULE     : floating.hpp
* DESCRIPTION: Floating point numbers
* COPYRIGHT  : (C) 2003--2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_FLOATING_HPP
#define __MML_FLOATING_HPP
#include "rational.hpp"
#include "string.hpp"
#include "type_props.hpp"
namespace mml {
#define TMPL_DEF template<typename V=std_floating>
#define TMPL template<typename V>
#define TMPL2 template<typename V,typename V2>
#define Floating floating<V>
#define Floating2 floating<V2>
#define Floating_rep floating_rep<V>
#define Rnd V::rnd()
TMPL class floating;
TMPL class floating_rep;

/******************************************************************************
* Different variants of floating types
******************************************************************************/

class std_floating {
public:
  static inline mp_rnd_t rnd () { return get_rounding_mode (); }
  static inline nat prec () {
    return get_precision (); }
  static inline nat prec (const mpfr_t& x) {
    return min (get_precision (), (nat) mpfr_get_prec (x)); }
  static inline nat prec (const mpfr_t& x, const mpfr_t& y) {
    return min (get_precision (),
		(nat) max (mpfr_get_prec (x), mpfr_get_prec (y))); }
};

class single_floating {
public:
  static inline mp_rnd_t rnd () { return get_rounding_mode (); }
  static inline nat prec () { return DEFAULT_LOW; }
  static inline nat prec (const mpfr_t& x) { (void) x; return DEFAULT_LOW; }
  static inline nat prec (const mpfr_t& x, const mpfr_t& y) {
    (void) x; (void) y; return DEFAULT_LOW; }
};

template<typename V, mp_rnd_t R>
class rnd_floating {
public:
  static inline mp_rnd_t rnd () { return R; }
  static inline nat prec () { return V::prec (); }
  static inline nat prec (const mpfr_t& x) { return V::prec (x); }
  static inline nat prec (const mpfr_t& x, const mpfr_t& y) {
    return V::prec (x,y); }
};

template<typename V, nat P>
class fixed_floating {
public:
  static inline mp_rnd_t rnd () { return V::rnd (); }
  static inline nat prec () { return P; }
  static inline nat prec (const mpfr_t& x) { (void) x; return P; }
  static inline nat prec (const mpfr_t& x, const mpfr_t& y) {
    (void) x; (void) y; return P; }
};

/******************************************************************************
* The floating type
******************************************************************************/

TMPL_DEF
class floating_rep REP_STRUCT {
private:
  mpfr_t x;

public:
  inline floating_rep () { mpfr_init2 (x, V::prec()); }
  inline floating_rep (nat prec) { mpfr_init2 (x, prec); }
  inline ~floating_rep () { mpfr_clear (x); }
  friend class Floating;
};

TMPL_DEF
class floating {
INDIRECT_PROTO_1 (floating, floating_rep, V)
public:
  // this should really be private, but we do not want to declare
  // the whole set of inline functions below twice
  inline floating (const mpfr_t& x):
    rep (new Floating_rep (V::prec (x))) {}
  inline floating (const mpfr_t& x, const mpfr_t& y):
    rep (new Floating_rep (V::prec (x, y))) {}
  inline mpfr_t& operator * () const { return rep->x; }

public:
  // constructors
  inline floating ():
    rep (new Floating_rep ()) {}
  inline floating (signed int i):
    rep (new Floating_rep ()) { mpfr_set_si (rep->x, i, Rnd); }
  inline floating (unsigned int i):
    rep (new Floating_rep ()) { mpfr_set_ui (rep->x, i, Rnd); }
  inline floating (signed short int i):
    rep (new Floating_rep ()) { mpfr_set_si (rep->x, i, Rnd); }
  inline floating (unsigned short int i):
    rep (new Floating_rep ()) { mpfr_set_ui (rep->x, i, Rnd); }
  inline floating (signed long int i):
    rep (new Floating_rep ()) { mpfr_set_si (rep->x, i, Rnd); }
  inline floating (unsigned long int i):
    rep (new Floating_rep ()) { mpfr_set_ui (rep->x, i, Rnd); }
  inline floating (float x2):
    rep (new Floating_rep ()) { mpfr_set_d (rep->x, x2, Rnd); }
  inline floating (double x2):
    rep (new Floating_rep ()) { mpfr_set_d (rep->x, x2, Rnd); }
  inline floating (const integer& i):
    rep (new Floating_rep ()) { mpfr_set_z (rep->x, *i, Rnd); }
  inline floating (const rational& x):
    rep (new Floating_rep ()) { mpfr_set_q (rep->x, *x, Rnd); }
  inline floating (char* s, int base= 0):
    rep (new Floating_rep ()) { mpfr_set_str (rep->x, s, base, Rnd); }
  inline floating (const string& s, int base= 0):
    rep (new Floating_rep ()) {
      char* r= as_charp (s);
      mpfr_set_str (rep->x, r, base, Rnd);
      free_charp (r); }
  friend class constant_helper<Floating >;
};
INDIRECT_IMPL_1 (floating, floating_rep, typename V, V)

TMPL
struct constant_helper<Floating > {
  inline static Floating nan () {
    Floating r; mpfr_init (*r); return r; }
  inline static Floating infinity () {
    Floating r; mpfr_set_inf (*r, 1); return r; }
  inline static Floating plus_infinity () {
    Floating r; mpfr_set_inf (*r, 1); return r; }
  inline static Floating minus_infinity () {
    Floating r; mpfr_set_inf (*r, -1); return r; }
  inline static Floating largest () {
    Floating r; mpfr_set_inf (*r, 1); mpfr_nextbelow (*r); return r; }
  inline static Floating smallest () {
    Floating r; mpfr_set_inf (*r, 1); mpfr_nextbelow (*r); return r; }
  inline static Floating log2 () {
    Floating r; mpfr_const_log2 (*r, Rnd); return r; }
  inline static Floating pi () {
    Floating r; mpfr_const_pi (*r, Rnd); return r; }
  inline static Floating euler () {
    Floating r; mpfr_const_euler (*r, Rnd); return r; }
  inline static Floating log2_D () {
    Floating r; mpfr_const_log2 (*r, GMP_RNDD); return r; }
  inline static Floating pi_D () {
    Floating r; mpfr_const_pi (*r, GMP_RNDD); return r; }
  inline static Floating euler_D () {
    Floating r; mpfr_const_euler (*r, GMP_RNDD); return r; }
  inline static Floating log2_U () {
    Floating r; mpfr_const_log2 (*r, GMP_RNDU); return r; }
  inline static Floating pi_U () {
    Floating r; mpfr_const_pi (*r, GMP_RNDU); return r; }
  inline static Floating euler_U () {
    Floating r; mpfr_const_euler (*r, GMP_RNDU); return r; }
};

/******************************************************************************
* Standard functions
******************************************************************************/

TMPL inline nat hash (const Floating& x) {
  __mpfr_struct &rep= (*x)[0];
  if (rep._mpfr_prec == 0) return 0;
  return ((nat) (rep._mpfr_d[0])) ^ ((nat) (rep._mpfr_exp << 2));
}

TMPL inline Floating copy (const Floating& x) {
  Floating r; mpfr_set (*r, *x, Rnd); return r; }

TMPL inline Floating copy (const Floating& x, nat prec) {
  Floating r (new Floating_rep (prec)); mpfr_set (*r, *x, Rnd); return r; }

/******************************************************************************
* Conversion functions
******************************************************************************/

TMPL string
as_string (const Floating& x, nat dd= 0) {
  if (x == 0) return "0";
  return trunc_digits (mpfr_to_string (*x), dd);
}

TMPL generic
as_generic (const Floating& g) {
  if (is_number (g) && g < 0) return (negate (as_string (-g)));
  else return as_string (g);
}

TMPL generic
as_generic (const Floating& g, nat dd) {
  if (is_number (g) && g < 0) return (negate (as_string (-g, dd)));
  else return as_string (g, dd);
}

TMPL generic
as_generic (const Floating& g1, const Floating& g2) {
  string s= make_range (as_string (g1), as_string (g2));
  if (s[0] == '-') return negate (as_generic (s (1, N(s))));
  else return as_generic (s);
}

TMPL generic
as_generic (const Floating& g1, const Floating& g2, nat dd) {
  string s= trunc_digits (make_range (as_string (g1), as_string (g2)), dd);
  if (s[0] == '-') return negate (as_generic (s (1, N(s))));
  else return as_generic (s);
}

TMPL ostream&
operator << (ostream& out, const Floating& x) {
  return out << as_string (x);
}

template<typename W, typename V> floating<W>
as_floating (const floating<V>& x) {
  floating<W> r;
  mpfr_set (*r, *x, W::rnd());
  return r;
}

template<typename W,typename V>
struct cast_helper<floating<W>,floating<V> > {
  static inline floating<W> cv (const floating<V>& x) {
    floating<W> r;
    mpfr_set (*r, *x, W::rnd());
    return r;
  }
};

TMPL inline int as_int (const Floating& x) {
  return mpfr_get_si (*x, Rnd); }
TMPL inline nat as_nat (const Floating& x) {
  return mpfr_get_ui (*x, Rnd); }
TMPL inline integer as_integer (const Floating& x) {
  integer r; mpfr_get_z (*r, *x, Rnd); return r; }

/******************************************************************************
* Predicates
******************************************************************************/

TMPL inline bool is_zero (const Floating& x) {
  return mpfr_sgn (*x) == 0; }
TMPL inline bool is_number (const Floating& x) {
  return mpfr_number_p (*x) != 0; }
TMPL inline bool is_infinity (const Floating& x) {
  return mpfr_inf_p (*x) != 0; }
TMPL inline bool is_nan (const Floating& x) {
  return mpfr_nan_p (*x) != 0; }
TMPL inline bool is_fuzz (const Floating& x) {
  (void) x; return false; }

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Floating operator - (const Floating& x1) {
  Floating r (*x1); mpfr_neg (*r, *x1, Rnd); return r; }
TMPL inline Floating operator + (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_add (*r, *x1, *x2, Rnd); return r; }
TMPL inline Floating operator - (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_sub (*r, *x1, *x2, Rnd); return r; }
TMPL inline Floating operator * (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_mul (*r, *x1, *x2, Rnd); return r; }
TMPL inline Floating operator / (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_div (*r, *x1, *x2, Rnd); return r; }
TMPL inline Floating operator << (const Floating& x1, int shift) {
  Floating r (*x1); mpfr_mul_2si (*r, *x1, shift, Rnd); return r; }
TMPL inline Floating operator >> (const Floating& x1, int shift) {
  Floating r (*x1); mpfr_mul_2si (*r, *x1, -shift, Rnd); return r; }
TMPL inline Floating& operator <<= (Floating& x1, int shift) {
  return x1= x1 << shift; }
TMPL inline Floating& operator >>= (Floating& x1, int shift) {
  return x1= x1 >> shift; }

ARITH_INT_SUGAR(TMPL,Floating,Floating);

/******************************************************************************
* Order related functions
******************************************************************************/

TMPL2 inline bool operator == (const Floating& x1, const Floating2& x2) {
  return mpfr_cmp (*x1, *x2) == 0; }
TMPL2 inline bool operator != (const Floating& x1, const Floating2& x2) {
  return mpfr_cmp (*x1, *x2) != 0; }
TMPL2 inline bool operator < (const Floating& x1, const Floating2& x2) {
  return mpfr_cmp (*x1, *x2) < 0; }
TMPL2 inline bool operator > (const Floating& x1, const Floating2& x2) {
  return mpfr_cmp (*x1, *x2) > 0; }
TMPL2 inline bool operator <= (const Floating& x1, const Floating2& x2) {
  return mpfr_cmp (*x1, *x2) <= 0; }
TMPL2 inline bool operator >= (const Floating& x1, const Floating2& x2) {
  return mpfr_cmp (*x1, *x2) >= 0; }

TMPL inline int sign (const Floating& x1) {
  int r= mpfr_sgn (*x1); return (r<0? -1: (r==0? 0: 1)); }
TMPL inline Floating abs (const Floating& x1) {
  Floating r (*x1); mpfr_abs (*r, *x1, Rnd); return r; }
TMPL inline Floating min (const Floating& x1, const Floating& x2) {
  return x1 <= x2? x1: x2; }
TMPL inline Floating max (const Floating& x1, const Floating& x2) {
  return x1 >= x2? x1: x2; }

TMPL inline Floating floor (const Floating& x1) {
  Floating r (*x1); mpfr_floor (*r, *x1); return r; }
TMPL inline Floating trunc (const Floating& x1) {
  Floating r (*x1); mpfr_trunc (*r, *x1); return r; }  
TMPL inline Floating ceil (const Floating& x1) {
  Floating r (*x1); mpfr_ceil (*r, *x1); return r; }
TMPL inline Floating round (const Floating& x1) {
  Floating r (*x1); mpfr_round (*r, *x1); return r; }

COMPARE_INT_SUGAR(TMPL,Floating,Floating);

/******************************************************************************
* Special functions
******************************************************************************/

TMPL inline Floating sqrt (const Floating& x1) {
  Floating r (*x1); mpfr_sqrt (*r, *x1, Rnd); return r; }
TMPL inline Floating cbrt (const Floating& x1) {
  Floating r (*x1); mpfr_cbrt (*r, *x1, Rnd); return r; }
TMPL inline Floating hypot (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_hypot (*r, *x1, *x2, Rnd); return r; }
TMPL inline Floating exp (const Floating& x1) {
  Floating r (*x1); mpfr_exp (*r, *x1, Rnd); return r; }
TMPL inline Floating exp2 (const Floating& x1) {
  Floating r (*x1); mpfr_exp2 (*r, *x1, Rnd); return r; }
TMPL inline Floating log (const Floating& x1) {
  Floating r (*x1); mpfr_log (*r, *x1, Rnd); return r; }
TMPL inline Floating log2 (const Floating& x1) {
  Floating r (*x1); mpfr_log2 (*r, *x1, Rnd); return r; }
TMPL inline Floating log10 (const Floating& x1) {
  Floating r (*x1); mpfr_log10 (*r, *x1, Rnd); return r; }
TMPL inline Floating pow (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_pow (*r, *x1, *x2, Rnd); return r; }
TMPL inline Floating cos (const Floating& x1) {
  Floating r (*x1); mpfr_cos (*r, *x1, Rnd); return r; }
TMPL inline Floating sin (const Floating& x1) {
  Floating r (*x1); mpfr_sin (*r, *x1, Rnd); return r; }
TMPL inline Floating tan (const Floating& x1) {
  Floating r (*x1); mpfr_tan (*r, *x1, Rnd); return r; }
TMPL inline Floating acos (const Floating& x1) {
  Floating r (*x1); mpfr_acos (*r, *x1, Rnd); return r; }
TMPL inline Floating asin (const Floating& x1) {
  Floating r (*x1); mpfr_asin (*r, *x1, Rnd); return r; }
TMPL inline Floating atan (const Floating& x1) {
  Floating r (*x1); mpfr_atan (*r, *x1, Rnd); return r; }
//TMPL inline Floating atan2 (const Floating& x1, const Floating& x2) {
//Floating r (*x1, *x2); mpfr_atan2 (*r, *x1, *x2, Rnd); return r; }
TMPL inline Floating cosh (const Floating& x1) {
  Floating r (*x1); mpfr_cosh (*r, *x1, Rnd); return r; }
TMPL inline Floating sinh (const Floating& x1) {
  Floating r (*x1); mpfr_sinh (*r, *x1, Rnd); return r; }
TMPL inline Floating tanh (const Floating& x1) {
  Floating r (*x1); mpfr_tanh (*r, *x1, Rnd); return r; }
TMPL inline Floating acosh (const Floating& x1) {
  Floating r (*x1); mpfr_acosh (*r, *x1, Rnd); return r; }
TMPL inline Floating asinh (const Floating& x1) {
  Floating r (*x1); mpfr_asinh (*r, *x1, Rnd); return r; }
TMPL inline Floating atanh (const Floating& x1) {
  Floating r (*x1); mpfr_atanh (*r, *x1, Rnd); return r; }
TMPL inline Floating gamma (const Floating& x1) {
  Floating r (*x1); mpfr_gamma (*r, *x1, Rnd); return r; }
TMPL inline Floating zeta (const Floating& x1) {
  Floating r (*x1); mpfr_zeta (*r, *x1, Rnd); return r; }
TMPL inline Floating erf (const Floating& x1) {
  Floating r (*x1); mpfr_erf (*r, *x1, Rnd); return r; }

/******************************************************************************
* Miscellaneous functions
******************************************************************************/

TMPL inline Floating next_above (const Floating& x1) {
  Floating r (*x1); mpfr_set (*r, *x1, Rnd);
  mpfr_nextabove (*r); return r; }
TMPL inline Floating next_below (const Floating& x1) {
  Floating r (*x1); mpfr_set (*r, *x1, Rnd);
  mpfr_nextbelow (*r); return r; }
TMPL inline nat precision (const Floating& x) {
  return (nat) mpfr_get_prec (*x); }
TMPL inline mp_exp_t exponent (const Floating& x) {
  return mpfr_safe_get_exp (*x); }
TMPL inline mp_exp_t step (const Floating& x) {
  return mpfr_safe_get_step (*x); }
TMPL inline Floating rnd_error (const Floating& x) {
  Floating r (*x);
  mpfr_mul_2si (*r, *x, 1-mpfr_get_prec (*x), GMP_RNDN);
  mpfr_abs (*r, *r, GMP_RNDN);
  return r; }

#undef TMPL_DEF
#undef TMPL
#undef TMPL2
#undef Floating
#undef Floating2
#undef Floating_rep
#undef Rnd
}
#endif // __MML_FLOATING_HPP
