
/******************************************************************************
* MODULE     : interval.hpp
* DESCRIPTION: Intervals
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_INTERVAL_HPP
#define __MML_INTERVAL_HPP
#include "rational.hpp"
#include "type_props.hpp"
namespace mml {
#define TMPL_DEF template<typename C,typename V=std_interval>
#define TMPL template<typename C,typename V>
#define Interval interval<C,V>
TMPL class interval;
TMPL inline C lower (const Interval& z);
TMPL inline C upper (const Interval& z);

/******************************************************************************
* Interval variants
******************************************************************************/

class std_interval {
public:
};

/******************************************************************************
* The interval type
******************************************************************************/

TMPL_DEF
class interval {
MML_ALLOCATORS;
private:
  C l;
  C r;

public:
  inline interval (): l (0), r (0) {}
  template<typename T> inline interval (const T& z):
    l (New_D(C) (z)), r (New_U(C) (z)) {}
  template<typename T> inline interval (const interval<T,V>& x):
    l (New_D(C) (lower (x))), r (New_U(C) (upper (x))) {}
  template<typename LT,typename UT> inline
  interval (const LT& l2, const UT& u2):
    l (New_D(C) (l2)), r (New_U(C) (u2)) {}
  inline interval (const C& x): l (x), r (x) {}
  inline interval (const C& l2, const C& r2): l (l2), r (r2) {}
  friend C lower LESSGTR (const Interval& x);
  friend C upper LESSGTR (const Interval& x);

  static inline Interval nan ();
  static inline Interval plus_infinity ();
  static inline Interval minus_infinity ();
  static inline Interval infinity ();
  static inline Interval fuzzy ();
  static inline Interval log2 ();
  static inline Interval pi ();
  static inline Interval euler ();

  inline Interval& operator += (const Interval& x);
  inline Interval& operator -= (const Interval& x);
  inline Interval& operator <<= (int shift);
  inline Interval& operator >>= (int shift);
};

STYPE_TO_TYPE(TMPL,center_type,Interval,C);
STYPE_TO_TYPE(TMPL,radius_type,Interval,C);

TMPL
struct constant_helper<Interval > {
  static inline Interval nan () {
    return Interval (Nan (C), Nan (C)); }
  static inline Interval plus_infinity () {
    return Interval (Plus_infinity (C), Plus_infinity (C)); }
  static inline Interval minus_infinity () {
    return Interval (Minus_infinity (C), Minus_infinity (C)); }
  static inline Interval infinity () {
    return Interval (Infinity (C), Infinity (C)); }
  static inline Interval fuzz () {
    return Interval (Minus_infinity (C), Plus_infinity (C)); }
  static inline Interval log2 () {
    return Interval (Log2_D (C), Log2_U (C)); }
  static inline Interval pi () {
    return Interval (Pi_D (C), Pi_U (C)); }
  static inline Interval euler () {
    return Interval (Euler_D (C), Euler_U (C)); }
};

TMPL
struct make_interval_helper<Interval,C> {
  static inline Interval val (const C& l, const C& r) {
    return Interval (l, r); }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C lower (const Interval& x) { return x.l; }
TMPL inline C upper (const Interval& x) { return x.r; }
TMPL inline C center (const Interval& x) {
  return (lower (x) + upper (x)) >> 1; }
TMPL inline C radius (const Interval& x) {
  if (is_infinity (x)) return 0;
  return add_U (sub_U (upper (x), lower (x)) >> 1, rnd_error (center (x))); }

TMPL inline bool is_exact_zero (const Interval& x) {
  return is_exact_zero (lower (x)) && is_exact_zero (upper (x)); }
TMPL inline bool is_number (const Interval& x) {
  return is_number (lower (x)) && is_number (upper (x)); }
TMPL inline bool is_infinity (const Interval& x) {
  return is_infinity (lower (x)) && upper (x) == lower (x); }
TMPL inline bool is_fuzz (const Interval& x) {
  return is_infinity (lower (x)) && is_infinity (upper (x)) &&
    lower (x) != upper (x); }
TMPL inline bool is_nan (const Interval& x) {
  return is_nan (lower (x)) || is_nan (upper (x)); }

TMPL inline Interval copy (const Interval& x, nat prec) {
  return Interval (copy (lower (x), prec), copy (upper (x), prec)); }
TMPL inline nat precision (const Interval& x) {
  return max (precision (lower (x)), precision (upper (x))); }
TMPL inline mp_exp_t expo (const Interval& x) {
  return max (exponent (lower (x)), exponent (upper (x))); }
TMPL inline mp_exp_t step (const Interval& x) {
  return step (upper (x) - lower (x)); }

TMPL inline nat
hash (const Interval& x) {
  nat h= hash (lower (x));
  return (h<<1) ^ (h<<3) ^ (h<<15) ^ (h>>29) ^ hash (upper (x));
}

TMPL generic
as_generic (const Interval& x) {
  if (is_nan (x)) return "NaN";
  if (lower (x) == 0 && upper (x) == 0) return 0;
  return as_generic (lower (x), upper (x));
  // return as_generic ("interval", lower (x), upper (x));
}

TMPL generic
as_generic (const Interval& x, nat dd) {
  if (is_nan (x)) return "NaN";
  if (lower (x) == 0 && upper (x) == 0) return 0;
  return as_generic (lower (x), upper (x), dd);
}

PRINT_SUGAR(TMPL,Interval)

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL inline bool
is_zero (const Interval& x) {
  return is_number (x) && sign (lower (x)) * sign (upper (x)) <= 0;
}

TMPL inline bool
is_non_zero (const Interval& x) {
  return !is_number (x) || sign (lower (x)) * sign (upper (x)) > 0;
}

TMPL inline bool
operator == (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return is_nan (x1) && is_nan (x2);
  return lower (x1) <= upper (x2) && lower (x2) <= upper (x1);
}

TMPL inline bool
operator != (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return !(is_nan (x1) && is_nan (x2));
  return lower (x1) > upper (x2) || lower (x2) > upper (x1);
}

TMPL inline bool
operator <= (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return false;
  return lower (x1) <= upper (x2);
}

TMPL inline bool
operator < (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return false;
  return lower (x2) > upper (x1);
}

TMPL inline bool
operator >= (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return false;
  return lower (x2) <= upper (x1);
}

TMPL inline bool
operator > (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return false;
  return lower (x1) > upper (x2);
}

COMPARE_INT_SUGAR(TMPL,Interval,Interval);

/******************************************************************************
* Other order related functions
******************************************************************************/

TMPL inline Interval
abs (const Interval& x) {
  if (is_nan (x)) return x;
  C r= max (abs (lower (x)), abs (upper (x)));
  if (sign (lower (x)) * sign (upper (x)) <= 0) return Interval (C(0), r);
  return Interval (min (abs (lower (x)), abs (upper (x))), r);
}

TMPL Interval
min (const Interval& x1, const Interval& x2) {
  return Interval (min (lower (x1), lower (x2)), min (upper (x1), upper (x2)));
}

TMPL Interval
max (const Interval& x1, const Interval& x2) {
  return Interval (max (lower (x1), lower (x2)), max (upper (x1), upper (x2)));
}

TMPL inline C abs_D (const Interval& x) { return lower (abs (x)); }
TMPL inline C abs_U (const Interval& x) { return upper (abs (x)); }
TMPL inline C bnd_D (const Interval& x) { return lower (x); }
TMPL inline C bnd_U (const Interval& x) { return upper (x); }

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Interval
operator - (const Interval& x) {
  return Interval (-upper (x), -lower (x));
}

TMPL inline Interval
operator + (const Interval& x1, const Interval& x2) {
  return Interval (add_D (lower (x1), lower (x2)),
		   add_U (upper (x1), upper (x2)));
}

TMPL inline Interval
operator - (const Interval& x1, const Interval& x2) {
  return Interval (sub_D (lower (x1), upper (x2)),
		   sub_U (upper (x1), lower (x2)));
}

TMPL Interval
operator * (const Interval& x1, const Interval& x2) {
  if (lower (x1) >= 0)
    return Interval (lower (x2) >= 0?
		     mul_D (lower (x1), lower (x2)):
		     mul_D (upper (x1), lower (x2)),
		     upper (x2) >= 0?
		     mul_U (upper (x1), upper (x2)):
		     mul_U (lower (x1), upper (x2)));
  else if (upper (x1) <= 0)
    return Interval (upper (x2) >= 0?
		     mul_D (lower (x1), upper (x2)):
		     mul_D (upper (x1), upper (x2)),
		     lower (x2) >= 0?
		     mul_U (upper (x1), lower (x2)):
		     mul_U (lower (x1), lower (x2)));
  else if (lower (x2) >= 0)
    return Interval (mul_D (lower (x1), upper (x2)),
		     mul_U (upper (x1), upper (x2)));
  else if (upper (x2) <= 0)
    return Interval (mul_D (upper (x1), lower (x2)),
		     mul_U (lower (x1), lower (x2)));
  else
    return Interval (min (mul_D (lower (x1), upper (x2)),
			  mul_D (upper (x1), lower (x2))),
		     max (mul_U (lower (x1), lower (x2)),
			  mul_U (upper (x1), upper (x2))));
}

TMPL inline Interval
square (const Interval& x) {
  if (lower (x) >= 0)
    return Interval (mul_D (lower (x), lower (x)),
		     mul_U (upper (x), upper (x)));
  else if (upper (x) <= 0)
    return Interval (mul_D (upper (x), upper (x)),
		     mul_U (lower (x), lower (x)));
  else return Interval (mul_D (lower (x), upper (x)),
			max (mul_U (lower (x), lower (x)),
			     mul_U (upper (x), upper (x))));
}

TMPL Interval
operator / (const Interval& x1, const Interval& x2) {
  if (is_zero (x2)) return Nan (Interval);
  else if (lower (x2) > 0)
    return Interval (lower (x1) > 0?
		     div_D (lower (x1), upper (x2)):
		     div_D (lower (x1), lower (x2)),
		     upper (x1) > 0?
		     div_U (upper (x1), lower (x2)):
		     div_U (upper (x1), upper (x2)));
  else if (upper (x2) < 0)
    return Interval (upper (x1) > 0?
		     div_D (upper (x1), upper (x2)):
		     div_D (upper (x1), lower (x2)),
		     lower (x1) > 0?
		     div_U (lower (x1), lower (x2)):
		     div_U (lower (x1), upper (x2)));
  else return Nan (Interval);
}

TMPL inline Interval
operator << (const Interval& x, int shift) {
  return Interval (lower (x) << shift, upper (x) << shift);
}

TMPL inline Interval
operator >> (const Interval& x, int shift) {
  return Interval (lower (x) >> shift, upper (x) >> shift);
}

ARITH_INT_SUGAR(TMPL,Interval,Interval);

/******************************************************************************
* In place arithmetic operations
******************************************************************************/

TMPL inline Interval&
Interval::operator += (const Interval& x) {
  l= add_D (l, lower (x));
  r= add_U (r, upper (x));
  return *this;
}

TMPL inline Interval&
Interval::operator -= (const Interval& x) {
  l= sub_D (l, upper (x));
  r= sub_U (r, lower (x));
  return *this;
}

TMPL inline Interval&
Interval::operator <<= (int shift) {
  l <<= shift;
  r <<= shift;
  return *this;
}

TMPL inline Interval&
Interval::operator >>= (int shift) {
  l >>= shift;
  r >>= shift;
  return *this;
}

/******************************************************************************
* Special functions
******************************************************************************/

TMPL Interval
sqrt (const Interval& x) {
  return Interval (sqrt_D (lower (x)), sqrt_U (upper (x)));
}

TMPL Interval
hypot (const Interval& x, const Interval& y) {
  Interval z= square (x) + square (y);
  if (z > 0) return sqrt (z);
  else return Interval (0, sqrt_U (upper (z)));
}

TMPL Interval
exp (const Interval& x) {
  return Interval (exp_D (lower (x)), exp_U (upper (x)));
}

TMPL Interval
log (const Interval& x) {
  return Interval (log_D (lower (x)), log_U (upper (x)));
}

TMPL inline Interval
pow (const Interval& x, const Interval& y) {
  return exp (y * log (x));
}

TMPL Interval
cos (const Interval& x) {
  if (sub_U (upper (x), lower (x)) >= C (3.14159)) return Interval (-1, 1);
  else {
    int s1= -sign (sin (lower (x)));
    int s2= -sign (sin (upper (x)));
    if (s1 >= 0 && s2 >= 0)
      return Interval (cos_D (lower (x)), cos_U (upper (x)));
    else if (s1 <= 0 && s2 <= 0)
      return Interval (cos_D (upper (x)), cos_D (lower (x)));
    else if (s1 >= 0)
      return Interval (min (cos_D (lower (x)), cos_D (upper (x))), C (1));
    else
      return Interval (C (-1), max (cos_D (lower (x)), cos_D (upper (x))));
  }
}

TMPL Interval
sin (const Interval& x) {
  if (sub_U (upper (x), lower (x)) >= C (3.14159)) return Interval (-1, 1);
  else {
    int s1= sign (cos (lower (x)));
    int s2= sign (cos (upper (x)));
    if (s1 >= 0 && s2 >= 0)
      return Interval (sin_D (lower (x)), sin_U (upper (x)));
    else if (s1 <= 0 && s2 <= 0)
      return Interval (sin_D (upper (x)), sin_D (lower (x)));
    else if (s1 >= 0)
      return Interval (min (sin_D (lower (x)), sin_D (upper (x))), C (1));
    else
      return Interval (C(-1), max (sin_D (lower (x)), sin_D (upper (x))));
  }
}

TMPL inline Interval
cosh (const Interval& x) {
  if (is_nan (x)) return Nan (Interval);
  else if (lower (x) >= 0)
    return Interval (cosh_D (lower (x)), cosh_U (upper (x)));
  else if (upper (x) <= 0)
    return Interval (cosh_D (upper (x)), cosh_U (lower (x)));
  else return Interval (1, max (cosh_U (lower (x)), cosh_U (upper (x))));
}

TMPL inline Interval
sinh (const Interval& x) {
  return Interval (sinh_D (lower (x)), sinh_U (upper (x)));
}

TMPL Interval
acos (const Interval& x) {
  return Interval (acos_D (upper (x)), acos_U (lower (x)));
}

TMPL Interval
asin (const Interval& x) {
  return Interval (asin_D (lower (x)), asin_U (upper (x)));
}

TMPL Interval
atan (const Interval& x) {
  return Interval (atan_D (lower (x)), atan_U (upper (x)));
}

TMPL Interval
atan2 (const Interval& y, const Interval& x) {
  if (sign (lower (x)) > 0) return atan (y/x);
  if (sign (lower (y)) > 0) return atan (-x/y) + (Pi (Interval) >> 1);
  if (sign (upper (y)) < 0) return atan (-x/y) - (Pi (Interval) >> 1);
  if (sign (upper (x)) < 0) return atan (y/x) + Pi (Interval);
  return Interval (-Pi_U (C), Pi_U (C));
}

#undef TMPL_DEF
#undef TMPL
#undef Interval
} // namespace mml
#endif // __MML_INTERVAL_HPP
