
/******************************************************************************
* MODULE     : karatsuba.hpp
* DESCRIPTION: low level pair/odd Karatsuba multiplication
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__KARATSUBA__HPP
#define __MML__KARATSUBA__HPP
#include "jet.hpp"

namespace mml {
#define TMPL template<class C>

/******************************************************************************
* Variant for Karatsuba multiplication
******************************************************************************/

class karatsuba_threshold {
public:
  static inline nat threshold () { return 2; }
};

template<typename V= karatsuba_threshold>
class karatsuba_multiply {};

/******************************************************************************
* "Karatsuba transforms"
******************************************************************************/

template<typename C> void
direct_fkt_step (C* dest, nat n) {
  // dest of length n on entry and 3n/2 on exit
  nat h= n >> 1;
  jet_half (dest + n, dest + 1, h);
  jet_half (dest + 1, dest + 2, h - 1);
  jet_add (dest + h, dest, dest + n, h);
}

template<typename C> void
direct_fkt (C* dest, nat n2, nat n3) {
  // dest of length n2 = 2^p on entry and n3 = 3^p on exit
  if (n2 == 2) {
    dest[2]= dest[1];
    dest[1] += dest[0];
  }
  else if (n2 == 4) {
    C p= dest[0] + dest[1];
    C q= dest[2] + dest[3];
    dest[6]= dest[2];
    dest[2]= dest[1];
    dest[8]= dest[3];
    dest[1]= p;
    dest[3]= dest[0] + dest[6];
    dest[4]= p+q;
    dest[5]= dest[2] + dest[8];
    dest[7]= q;
  }
  else while (n2 != n3) {
    direct_fkt_step (dest, n2);
    n2= n2 + (n2 >> 1);
  }
}

template<typename C> void
inverse_fkt_step (C* dest, nat n, nat l) {
  // dest of length n = 3 k l on entry for some integer k
  // dest of length (2 l + 1) k on exit
  nat t= n/3, k= t/l, ll= (l << 1) + 1;
  jet_sub (dest + t, dest, t);
  jet_sub (dest + t, dest + (t << 1), t);
  INLINE_ARRAY (C, buf, n);
  jet_copy (buf, dest, n);
  C* src= buf;
  if (l == 1) {
    jet_triple (dest, buf, t);
    jet_triple (dest + 1, buf + t, t);
    jet_triple (dest + 2, buf + (t << 1), t);
  }
  else while (k != 0) {
    jet_double (dest, src, l);
    jet_double (dest + 1, src + t, l);
    jet_double_add (dest + 2, src + (t << 1), l-1);
    dest[l<<1]= src[(t<<1) + (l-1)];
    dest += ll; src += l; k--;
  }
  DELETE_ARRAY (C, buf, n);
}

template<typename C> void
inverse_fkt (C* dest, nat n3, nat n2) {
  // dest of length n3 = 3^p on entry and 2 n2 - 1 on exit with n2 = 2 ^ p
  if (n2 == 2)
    dest[1] -= (dest[0] + dest[2]);
  else if (n2 == 4) {
    dest[3] -= (dest[0] + dest[6]);
    dest[4] -= (dest[1] + dest[7]);
    dest[5] -= (dest[2] + dest[8]);
    C aux= dest[1]; dest[1]= dest[3]; dest[3]= aux;
    aux= dest[2]; dest[2]= dest[6]; dest[6]= aux;
    aux= dest[5]; dest[5]= dest[7]; dest[7]= aux;
    dest[3] -= (dest[0] + dest[6]);
    dest[4] -= (dest[1] + dest[7]);
    dest[5] -= (dest[2] + dest[8]);
    aux= dest[1];
    dest[1]= dest[3];
    dest[3]= dest[4];
    dest[4]= dest[2] + dest[7];
    dest[2]= aux + dest[6];
    dest[6]= dest[8];
  }
  else {
    nat n= n3, l= 1;
    while (n != l) {
      inverse_fkt_step (dest, n, l);
      nat k= n / ((l << 1) + l);
      l= (l << 1) + 1;
      n= k * l;
    }
  }
}

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename C, typename V>
class multiplier<C,karatsuba_multiply<V>,C > {
  static void multiply (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    if (n2 < V::threshold ()) {
      jet_mul<C,naive_multiply> (dest, s1, s2, n1, n2); return; }
    nat p1= n1 >> 1, p2= n2 >> 1, P= p1+p2-1, spc= 3*(p1+p2);
    INLINE_ARRAY (C, buf, spc);
    C* low1= dest;
    C* low2= low1 + p1;
    C* mid1= buf;
    C* mid2= buf + p1;
    C* hi1 = mid2 + p2;
    C* hi2 = hi1 + p1;
    C* Low = mid1;
    C* Mid = hi1;
    C* Hi  = Mid + p1 + p2;
    jet_half (low1, s1, p1);
    jet_half (hi1 , s1+1, p1);
    jet_add  (mid1, low1, hi1, p1);
    jet_half (low2, s2, p2);
    jet_half (hi2 , s2+1, p2);
    jet_add  (mid2, low2, hi2, p2);
    multiply (Hi , hi1 , hi2 , p1, p2);
    multiply (Mid, mid1, mid2, p1, p2);
    multiply (Low, low1, low2, p1, p2);
    jet_sub  (Mid, Low, P);
    jet_sub  (Mid, Hi , P);
    jet_add  (Low+1, Hi, P-1);
    Low[P]= Hi[P-1];
    jet_double (dest, Low, P+1);
    jet_double (dest+1, Mid, P);
    DELETE_ARRAY (C, buf, spc);

    if ((n1 & 1) != 0) {
      dest[(P<<1)+1]= C(0);
      jet_mul_add (dest + (p1<<1), s2, s1[p1<<1], p2<<1);
    }
    if ((n2 & 1) != 0) {
      dest[n1+n2-2]= C(0);
      jet_mul_add (dest + (p2<<1), s1, s2[p2<<1], n1);
    }
  }

public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    if (n1 == 0 || n2 == 0) return;
    if (n1 < n2) multiply (dest, s2, s1, n2, n1);
    else multiply (dest, s1, s2, n1, n2);
  }
};

/******************************************************************************
* Division
******************************************************************************/

template<typename C, typename V>
class divisor<C,karatsuba_multiply<V> > {
public:
  static void quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
    return jet_dicho_quo_rem<C,karatsuba_multiply<V> >
             (dest, s1, s2, n1, n2, V::threshold ());
  }
};

#undef TMPL
} // namespace mml
#endif //__MML__KARATSUBA__HPP
