
/******************************************************************************
* MODULE     : segment.hpp
* DESCRIPTION: header for low level vectorial computations
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__SEGMENT__HPP
#define __MML__SEGMENT__HPP
#include "base.hpp"
#include "operators.hpp"

namespace mml {
#define TMPL template<typename C>

#define INLINE_ARRAY(TYPE,name,len) TYPE name[len]
#define DELETE_ARRAY(TYPE,name,len)

/******************************************************************************
* Naive multiplication (vectors, matrices, polynomials and series)
******************************************************************************/

class naive_multiply {};

/******************************************************************************
* Abstract low level vector routines on segments
******************************************************************************/

template<typename Op, typename C> inline void
segment_unary (C* dest, const C* s, nat n) {
  while (n != 0) {
    *dest= Op::op (*s);
    dest++; s++; n--;
  }
}

template<typename Op, typename R, typename C> inline void
segment_cast_unary (R* dest, const C* s, nat n) {
  while (n != 0) {
    *dest= Op::template cast_helper<R>::cast_op (*s);
    dest++; s++; n--;
  }
}

template<typename Op, typename C> inline void
segment_in_place_unary (C* dest, nat n) {
  while (n != 0) {
    Op::in_place (*dest);
    dest++; n--;
  }
}

template<typename Op, typename C> inline void
segment_binary (C* dest, const C* s1, const C* s2, nat n) {
  while (n != 0) {
    *dest = Op::op (*s1, *s2);
    dest++; s1++; s2++; n--;
  }
}

template<typename Op, typename C> inline void
segment_in_place_binary (C* dest, const C* s, nat n) {
  while (n != 0) {
    *dest = Op::in_place (*dest, *s);
    dest++; s++; n--;
  }
}

template<typename Op, typename C, typename X> inline void
segment_extrary (C* dest, const C* s, const X& x, nat n) {
  while (n != 0) {
    *dest = Op::op (*s, x);
    dest++; s++; n--;
  }
}

template<typename Op, typename R, typename C, typename X> inline void
segment_cast_extrary (R* dest, const C* s, const X& x, nat n) {
  while (n != 0) {
    *dest= Op::template cast_helper<R>::cast_op (*s, x);
    dest++; s++; n--;
  }
}

template<typename Op, typename C, typename X> inline void
segment_in_place_extrary (C* dest, const X& x, nat n) {
  while (n != 0) {
    Op::in_place (*dest, x);
    dest++; n--;
  }
}

template<typename Op, typename C> inline bool
segment_test (const C* s1, const C* s2, nat n) {
  while (n != 0) {
    if (Op::not_op (*s1, *s2)) return false;
    s1++; s2++; n--;
  }
  return true;
}

template<typename Op, typename X, typename C> inline bool
segment_test_extra (const C* s, const X& x, nat n) {
  while (n != 0) {
    if (Op::not_op (*s, x)) return false;
    s++; n--;
  }
  return true;
}

template<typename Op, typename C> inline C
segment_big (const C* s, nat n) {
  if (n == 0) return Op::template neutral<C> ();
  C r= s[0];
  for (nat i=1; i<n; i++)
    Op::in_place (r, s[i]);
  return r;
}

/******************************************************************************
* Instantiations
******************************************************************************/

TMPL void segment_neg (C* dest, const C* s, nat n) {
  segment_unary<neg_op> (dest, s, n); }
TMPL void segment_neg (C* dest, nat n) {
  segment_in_place_unary<neg_op> (dest, n); }
TMPL void segment_add (C* dest, const C* s1, const C* s2, nat n) {
  segment_binary<add_op> (dest, s1, s2, n); }
TMPL void segment_sub (C* dest, const C* s1, const C* s2, nat n) {
  segment_binary<sub_op> (dest, s1, s2, n); }
TMPL void segment_mul (C* dest, const C* s1, const C* s2, nat n) {
  segment_binary<mul_op> (dest, s1, s2, n); }
TMPL void segment_div (C* dest, const C* s1, const C* s2, nat n) {
  segment_binary<div_op> (dest, s1, s2, n); }
TMPL void segment_mul (C* dest, const C* s, const C& c, nat n) {
  segment_extrary<mul_op> (dest, s, c, n); }
TMPL void segment_div (C* dest, const C* s, const C& c, nat n) {
  segment_extrary<div_op> (dest, s, c, n); }
TMPL void segment_quo (C* dest, const C* s, const C& c, nat n) {
  segment_extrary<quo_op> (dest, s, c, n); }
TMPL void segment_rem (C* dest, const C* s, const C& c, nat n) {
  segment_extrary<rem_op> (dest, s, c, n); }
TMPL void segment_copy (C* dest, const C* s, nat n) {
  segment_in_place_binary<copy_op> (dest, s, n); }
TMPL void segment_add (C* dest, const C* s, nat n) {
  segment_in_place_binary<add_op> (dest, s, n); }
TMPL void segment_sub (C* dest, const C* s, nat n) {
  segment_in_place_binary<sub_op> (dest, s, n); }
TMPL void segment_mul (C* dest, const C* s, nat n) {
  segment_in_place_binary<mul_op> (dest, s, n); }
TMPL void segment_div (C* dest, const C* s, nat n) {
  segment_in_place_binary<div_op> (dest, s, n); }
TMPL void segment_mul (C* dest, const C& c, nat n) {
  segment_in_place_extrary<mul_op> (dest, c, n); }
TMPL void segment_div (C* dest, const C& c, nat n) {
  segment_in_place_extrary<div_op> (dest, c, n); }
TMPL void segment_quo (C* dest, const C& c, nat n) {
  segment_in_place_extrary<quo_op> (dest, c, n); }
TMPL void segment_rem (C* dest, const C& c, nat n) {
  segment_in_place_extrary<rem_op> (dest, c, n); }
TMPL bool segment_eq (const C* s1, const C* s2, nat n) {
  return segment_test<equal_op> (s1, s2, n); }
TMPL bool segment_lesseq (const C* s1, const C* s2, nat n) {
  return segment_test<lesseq_op> (s1, s2, n); }
TMPL bool segment_gtreq (const C* s1, const C* s2, nat n) {
  return segment_test<gtreq_op> (s1, s2, n); }
TMPL bool segment_eq (const C* s, const C& c, nat n) {
  return segment_test_extra<equal_op> (s, c, n); }
TMPL bool segment_lesseq (const C* s, const C& c, nat n) {
  return segment_test_extra<lesseq_op> (s, c, n); }
TMPL bool segment_gtreq (const C* s, const C& c, nat n) {
  return segment_test_extra<gtreq_op> (s, c, n); }

/******************************************************************************
* Further routines
******************************************************************************/

TMPL void
segment_set (C* dest, const C& c, nat n) {
  while (n != 0) {
    *dest = c;
    dest++; n--;
  }
}

TMPL void
segment_clear (C* dest, nat n) {
  while (n != 0) {
    *dest = C(0);
    dest++; n--;
  }
}

template<typename C, typename T> void
segment_cast (C* dest, const T* s, nat n) {
  while (n != 0) {
    (*dest) = (*s);
    dest++; s++; n--;
  }
}

template<typename C, typename K> void
segment_mul_add (C* dest, const C* s, const K& c, nat n) {
  while (n != 0) {
    *dest += (*s) * c;
    dest++; s++; n--;
  }
}

template<typename C, typename K> void
segment_mul_add (C* dest, const K& c, const C* s, nat n) {
  while (n != 0) {
    *dest += c * (*s);
    dest++; s++; n--;
  }
}

template<typename C> void
segment_mul_add (C* dest, const C* s1, const C* s2, nat n) {
  while (n != 0) {
    *dest += (*s1) * (*s2);
    dest++; s1++; s2++; n--;
  }
}

TMPL C
segment_inn_prod (C* s1, const C* s2, nat n) {
  C r (0);
  while (n != 0) {
    r += (*s1) * (*s2);
    s1++; s2++;
  }
  return r;
}

TMPL void
segment_print (ostream& out, const C* s, nat n) {
  if (n == 0) out << "[]";
  else {
    out << "[ " << s[0];
    for (nat i=1; i<n; i++)
      out << ", " << s[i];
    out << " ]";
  }
}

#undef TMPL
} // namespace mml
#endif //__MML__SEGMENT__HPP
