
/******************************************************************************
* MODULE     : linear_series.hpp
* DESCRIPTION: Vectors of univariate power series
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_LINEAR_SERIES_HPP
#define __MML_LINEAR_SERIES_HPP
#include "series.hpp"
#include "dense_vector.hpp"
#include "dense_matrix.hpp"
namespace mml {
#define TMPL_DEF template<typename C, typename V, typename W>
#define TMPL template<typename C, typename V, typename W>
#define Series series<C,V>
#define Series_rep series_rep<C,V>

/******************************************************************************
* Vectors of series
******************************************************************************/

#define Series_vector series<vector<C,W>,V>
#define Vector_series vector<series<C,V>,W>
STYPE_TO_TYPE(TMPL,as_vector_type,Series_vector,Vector_series);
#undef Vector_series
#undef Series_vector

TMPL
class vector_access_series_rep : public Series_rep {
  series<vector<C,W>,V> f;
  nat i;
public:
  vector_access_series_rep (const series<vector<C,W>,V>& f2, nat i2):
    f (f2), i (i2) {}
  C next () { return f[this->n][i]; }
  generic As_generic (const generic& z) const {
    return access (as_generic (f, z), as_generic (i)); }
};

TMPL Series
access (const series<vector<C,W>,V>& f, nat i) {
  return (Series_rep*) new vector_access_series_rep<C,V,W> (f, i);
}

TMPL vector<Series,W>
as_vector (const series<vector<C,W>,V>& f) {
  nat n= dim (f[0]);
  vector<Series,W> v (Series (0), n);
  for (nat i=0; i<n; i++)
    v (i)= access (f, i);
  return v;
}

TMPL generic
as_generic (const vector<Series,W>& v, const generic& z) {
  generic g;
  nat i, n= dim (v);
  vector<generic,W> r (g, n);
  for (i=0; i<n; i++)
    r (i)= as_generic (v[i], z);
  return as_generic (r);
}

TMPL
class vector_series_rep : public series_rep<vector<C,W>,V> {
  vector<Series,W> v;
public:
  vector_series_rep (const vector<Series,W>& v2): v (v2) {}
  vector<C,W> next () {
    nat i, n= dim (v);
    vector<C,W> coeff (C (0), n);
    for (i=0; i<n; i++)
      coeff (i)= v[i][this->n];
    return coeff; }
  generic As_generic (const generic& z) const {
    return as_generic (v, z); }
};

TMPL series<vector<C,W>,V>
as_series (const vector<Series,W>& v) {
  return (series_rep<vector<C,W>,V>*) new vector_series_rep<C,V,W> (v);
}

TMPL series<vector<C,W>,V>
from_vector (const vector<Series,W>& v) {
  return (series_rep<vector<C,W>,V>*) new vector_series_rep<C,V,W> (v);
}

/******************************************************************************
* Trigonometric functions and linear differential equations
******************************************************************************/

TMPL series<vector<C,W>,V>
trig (const series<vector<C,W>,V>& f) {
  return unary_implicit_series<trig_op> (f);
}

template<typename C,typename V> vector<Series >
cos_sin (const Series& f) {
  return as_vector (trig (as_series (vec (f))));
}

template<typename C,typename V> Series
cos (const Series& f) {
  return unary_implicit_series<cos_op> (f);
}

template<typename C,typename V> Series
sin (const Series& f) {
  return unary_implicit_series<sin_op> (f);
}

template<typename C,typename V> Series
tan (const Series& f) {
  return unary_implicit_series<tan_op> (f);
}

TMPL series<vector<C,W>,V>
solve_vector_lde_init (const series<vector<C,W>,V>& f, const vector<C,W>& c) {
  return unary_implicit_series<solve_vector_lde_op> (f, c);
}

TMPL vector<Series,W>
solve_lde_init (const vector<Series,W>& f, const vector<C,W>& c) {
  return as_vector (solve_vector_lde_init (as_series (f), c));
}

TMPL Series // FIXME: obsolete
solve_lde (const vector<Series,W>& f, const vector<C,W>& c) {
  return solve_lde_init (f, c) [0];
}

/******************************************************************************
* Matrices of series
******************************************************************************/

#define Series_matrix series<matrix<C,W>,V>
#define Matrix_series matrix<series<C,V>,W>
STYPE_TO_TYPE(TMPL,as_matrix_type,Series_matrix,Matrix_series);
#undef Matrix_series
#undef Series_matrix

TMPL
class matrix_access_series_rep : public Series_rep {
  series<matrix<C,W>,V> f;
  nat i, j;
public:
  matrix_access_series_rep (const series<matrix<C,W>,V>& f2, nat i2, nat j2):
    f (f2), i (i2), j (j2) {}
  C next () { return f[this->n](i,j); }
  generic As_generic (const generic& z) const {
    return access (as_generic (f, z), as_generic (i), as_generic (j)); }
};

TMPL Series
access (const series<matrix<C,W>,V>& f, nat i, nat j) {
  return (Series_rep*) new matrix_access_series_rep<C,V,W> (f, i, j);
}

TMPL matrix<Series,W>
as_matrix (const series<matrix<C,W>,V>& f) {
  nat nr= rows (f[0]), nc= cols (f[0]);
  matrix<Series,W> m (Series (0), nr, nc);
  for (nat i=0; i<nr; i++)
    for (nat j=0; j<nc; j++)
      m.ref (i, j)= access (f, i, j);
  return m;
}

TMPL generic
as_generic (const matrix<Series,W>& m, const generic& z) {
  generic g;
  nat i, j, nr= rows (m), nc= cols (m);
  matrix<generic,W> r (g, nr, nc);
  for (i=0; i<nr; i++)
    for (j=0; j<nc; j++)
      r.ref (i, j)= as_generic (m (i, j), z);
  return as_generic (r);
}

TMPL
class matrix_series_rep : public series_rep<matrix<C,W>,V> {
  matrix<Series,W> m;
public:
  matrix_series_rep (const matrix<Series,W>& m2): m (m2) {}
  matrix<C,W> next () {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<C,W> coeff (C (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	coeff.ref (i, j)= m (i, j) [this->n];
    return coeff; }
  generic As_generic (const generic& z) const {
    return as_generic (m, z); }
};

TMPL series<matrix<C,W>,V>
as_series (const matrix<Series,W>& m) {
  return (series_rep<matrix<C,W>,V>*) new matrix_series_rep<C,V,W> (m);
}

TMPL series<matrix<C,W>,V>
from_matrix (const matrix<Series,W>& m) {
  return (series_rep<matrix<C,W>,V>*) new matrix_series_rep<C,V,W> (m);
}

/******************************************************************************
* Linear differential matrix equations
******************************************************************************/

TMPL series<matrix<C,W>,V>
solve_matrix_lde_init (const series<matrix<C,W>,V>& f, const matrix<C,W>& c) {
  return unary_implicit_series<solve_matrix_lde_op> (f, c);
}

TMPL matrix<Series,W>
solve_lde (const matrix<Series,W>& f) {
  assert (is_square_matrix (f));
  matrix<C,W> c (C(1), rows (f), cols (f));
  return as_matrix (solve_matrix_lde_init (as_series (f), c));
}

TMPL matrix<Series,W>
solve_lde_init (const matrix<Series,W>& f, const matrix<C,W>& c) {
  assert (is_square_matrix (f) && rows (f) == rows (c));
  return as_matrix (solve_matrix_lde_init (as_series (f), c));
}

#undef TMPL_DEF
#undef TMPL
#undef Series
#undef Series_rep
} // namespace mml
#endif // __MML_LINEAR_SERIES_HPP
