
/******************************************************************************
* MODULE     : symbolic_calculus.hpp
* DESCRIPTION: Basic calculus with symbolic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_CALCULUS_HPP
#define __SYMBOLIC_CALCULUS_HPP
#include "symbolic.hpp"
namespace mml {
#define TMPL template<typename C,typename V>
#define Symbolic symbolic<C,V>
#define Table table<Symbolic,Symbolic >
#define Vector vector<Symbolic >
#define Iterator iterator<Symbolic >

/******************************************************************************
* Exp-log functions
******************************************************************************/

TMPL Symbolic
sqrt (const Symbolic& x) {
  return pow (x, Symbolic (C(1)/C(2)));
}

TMPL Symbolic
exp (const Symbolic& x) {
  return pow (Symbolic (MML_E), x);
}

TMPL Symbolic
log (const Symbolic& x) {
  if (x == 1) return 0;
  if (x == Symbolic (MML_E)) return 1;
  if (is_sum (x)) {
    Table t= get_sum (x);
    if (N(t) == 1) {
      Symbolic y = *entries (t);
      Symbolic sc= t[y];
      return log (sc) + log (y);
    }
    else return Symbolic (MML_LOG) (x);
  }
  if (is_product (x)) {
    Table t= get_product (x);
    Vector v;
    for (Iterator it= entries (t); busy (it); ++it)
      v << t[*it] * log (*it);
    return sum (v);
  }
  return Symbolic (MML_LOG) (x);
}

/******************************************************************************
* Trigonometric functions
******************************************************************************/

TMPL Symbolic
sin (const Symbolic& x) {
  if (x == 0) return 0;
  return Symbolic (MML_SIN) (x);
}

TMPL Symbolic
cos (const Symbolic& x) {
  if (x == 0) return 1;
  return Symbolic (MML_COS) (x);
}

TMPL Symbolic
tan (const Symbolic& x) {
  if (x == 0) return 0;
  return Symbolic (MML_TAN) (x);
}

TMPL Symbolic
asin (const Symbolic& x) {
  if (x == 0) return 0;
  return Symbolic (MML_ARCSIN) (x);
}

TMPL Symbolic
acos (const Symbolic& x) {
  if (x == 1) return 0;
  return Symbolic (MML_ARCCOS) (x);
}

TMPL Symbolic
atan (const Symbolic& x) {
  if (x == 0) return 0;
  return Symbolic (MML_ARCTAN) (x);
}

/******************************************************************************
* Differentiation
******************************************************************************/

TMPL bool
depends (const Symbolic& x, const Symbolic& y) {
  if (is_scalar (x)) return false;
  if (x == y) return true;
  if (is_atom (x)) return false;
  if (is_sum (x)) {
    Table t= get_sum (x);
    for (Iterator it= entries (t); busy (it); ++it)
      if (depends (*it, y) || depends (t[*it], y)) return true;
    return false;
  }
  if (is_product (x)) {
    Table t= get_product (x);
    for (Iterator it= entries (t); busy (it); ++it)
      if (depends (*it, y) || depends (t[*it], y)) return true;
    return false;
  }
  if (is_compound (x)) {
    for (nat i=0; i<N(x); i++)
      if (depends (x[i], y)) return true;
    return false;
  }
  return false;
}

TMPL Symbolic
derive (const Symbolic& x, const Symbolic& y) {
  if (is_scalar (x)) return 0;
  if (x == y) return 1;
  if (is_atom (x)) return 0;
  if (is_sum (x)) {
    Table t= get_sum (x);
    Vector v;
    for (Iterator it= entries (t); busy (it); ++it)
      v << t[*it] * derive (*it, y);
    return sum (v);
  }
  if (V::general_commute) {
    if (is_product (x)) {
      Table t= get_product (x);
      Vector v;
      for (Iterator it= entries (t); busy (it); ++it) {
	Symbolic h= *it, sc= t[h];
	v << derive (sc, y) * log (h) + sc * derive (h, y) / h;
      }
      return sum (v) * x;
    }
    if (is_func (x, MML_LOG, 1))
      return derive (x[1], y) / x[1];
    if (is_func (x, MML_SIN, 1))
      return derive (x[1], y) * cos (x[1]);
    if (is_func (x, MML_COS, 1))
      return -derive (x[1], y) * sin (x[1]);
    if (is_func (x, MML_TAN, 1))
      return derive (x[1], y) / pow (cos (x[1]), 2);
    if (is_func (x, MML_ARCSIN, 1))
      return derive (x[1], y) / sqrt (1 - x[1] * x[1]);
    if (is_func (x, MML_ARCCOS, 1))
      return -derive (x[1], y) / sqrt (1 - x[1] * x[1]);
    if (is_func (x, MML_ARCTAN, 1))
      return derive (x[1], y) / (x[1] * x[1] + 1);
    if (is_compound (x)) {
      if (N(x) == 2)
	return derive (x[1], y) * (Symbolic (MML_PRIME) (x[0])) (x[1]);
      Vector v;
      for (nat i=1; i<N(x); i++) {
	Vector f= copy (as_vector (x));
	f(0)= (Symbolic (MML_ACCESS) ("D", i)) (x[0]);
	v << derive (x[i], y) * Symbolic (f);
      }
      return sum (v);
    }
  }
  else {
    if (is_func (x, MML_TIMES, 2))
      return derive (x[1], y) * x[2] + x[1] * derive (x[2], y);
    if (is_func (x, MML_POWER, 2) && x[1] == y &&
	is_scalar (x[2]) && V::scalar_commute)
      return x[2] * pow (y, x[2] - 1);
    if (!depends (x, y)) return 0;
    return (Symbolic (MML_ACCESS) ("D", y)) (x);
  }
  return 0;
}

TMPL Symbolic
derive (const Symbolic& s1, const Symbolic& s2, nat n) {
  if (n == 0) return s1;
  else return derive (derive (s1, s2, n-1), s2);
}

#undef TMPL
#undef Symbolic
#undef Table
#undef Vector
#undef Iterator
} // namespace mml
#endif // __SYMBOLIC_CALCULUS_HPP
